from collections import deque
from decimal import Decimal
from typing import Deque, Literal, Optional

from barbthroat.math import minmax
from barbthroat.models import Candle

from .changed_filter import ChangedFilter
from .simple_moving_average import SimpleMovingAverage

Advice = Literal["Long", "Short", "Liquidate"]


class FourWeekRule:
    """
    The weekly rule, in its simplest form, buys when prices reach a new four-week high and sells
    when prices reach a new four-week low.
    """

    _changed_filter: ChangedFilter
    _moving_average: SimpleMovingAverage
    _prices: Deque[Decimal]

    def __init__(
        self,
        period: int = 28,
        # Moving average period is normally half the period.
        moving_average_period: int = 14,
    ) -> None:
        self._changed_filter = ChangedFilter()
        self._moving_average = SimpleMovingAverage(period=moving_average_period)
        self._prices = deque(maxlen=period)

    @property
    def value(self) -> Optional[Advice]:
        return self._changed_filter.value

    def update(self, candle: Candle) -> Optional[Advice]:
        price = candle.close

        self._moving_average.update(price)
        self._prices.append(price)

        value = None
        lowest, highest = minmax(self._prices)
        if price >= highest:
            value = "Long"
        elif price <= lowest:
            value = "Short"
        elif (
            self._changed_filter.prevailing_value == "Long" and price <= self._moving_average.value
        ) or (
            self._changed_filter.prevailing_value == "Short"
            and price >= self._moving_average.value
        ):
            value = "Liquidate"
        return self._changed_filter.update(value)



from decimal import Decimal
from typing import Iterable, Tuple


def minmax(values: Iterable[Decimal]) -> Tuple[Decimal, Decimal]:
    min_ = Decimal("Inf")
    max_ = Decimal("-Inf")
    for value in values:
        min_ = min(min_, value)
        max_ = max(max_, value)
    return min_, max_


def floor_multiple(value: int, multiple: int) -> int:
    return value - (value % multiple)


def floor_multiple_offset(value: int, multiple: int, offset: int) -> int:
    return floor_multiple(value - offset, multiple) + offset



from __future__ import annotations

import asyncio
import logging
from contextlib import asynccontextmanager
from decimal import Decimal
from functools import partial
from types import TracebackType
from typing import Any, AsyncIterable, AsyncIterator, Dict, Optional

import aiohttp
import simplejson as json
from aiolimiter import AsyncLimiter
from multidict import CIMultiDictProxy, istr

from barbthroat import time
from barbthroat.http import ClientResponse, ClientSession, connect_refreshing_stream
from barbthroat.models import Candle, ConnectorException, SavingsProduct, Trade

_log = logging.getLogger(__name__)

_BASE_API_URL = "https://api.binance.com"
_BASE_WS_URL = "wss://stream.binance.com:9443"

_SEC_NONE = 0  # Endpoint can be accessed freely.
_SEC_USER_DATA = 2  # Endpoint requires sending a valid API-Key and signature.

_ERR_TOO_MANY_REQUESTS = -1003


class BinancePaperTradeConnector:
    # Capabilities.
    can_stream_historical_candles: bool = True
    can_stream_candles: bool = True

    def __init__(self) -> None:
        self._session = ClientSession(name=type(self).__name__)

        # Rate limiters.
        x = 1.5  # We use this factor to be on the safe side and not use up the entire bucket.
        self._reqs_per_min_limiter = AsyncLimiter(1200, 60 * x)
        self._raw_reqs_limiter = AsyncLimiter(5000, 300 * x)

    async def __aenter__(self) -> BinancePaperTradeConnector:
        await self._session.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        await self._session.__aexit__(exc_type, exc, tb)

    # Candles.

    def list_candle_intervals(self) -> list[int]:
        return [
            60000,  # 1m
            180000,  # 3m
            300000,  # 5m
            900000,  # 15m
            1800000,  # 30m
            3600000,  # 1h
            7200000,  # 2h
            14400000,  # 4h
            21600000,  # 6h
            28800000,  # 8h
            43200000,  # 12h
            86400000,  # 1d
            259200000,  # 3d
            604800000,  # 1w
            2629746000,  # 1M
        ]

    async def stream_historical_candles(
        self,
        trading_pair: str,
        interval: int,
        start: int,
        end: int,
    ) -> AsyncIterable[Candle]:
        limit = 1000  # Max possible candles per request.
        binance_interval = time.format_interval(interval)
        binance_trading_pair = _to_binance_trading_pair(trading_pair)
        binance_start = start
        binance_end = end - 1
        while True:
            if binance_start > binance_end:
                return
            content = await self._api_request(
                method="GET",
                url="/api/v3/klines",
                data={
                    "symbol": binance_trading_pair,
                    "interval": binance_interval,
                    "startTime": binance_start,
                    "endTime": binance_end,
                    "limit": limit,
                },
            )
            for c in content:
                # Binance can return bad candles where the time does not fall within the requested
                # interval. For example, the second candle of the following query has bad time:
                # https://api.binance.com/api/v3/klines?symbol=ETHBTC&interval=4h&limit=10&startTime=1529971200000&endTime=1530000000000
                candle_time = c[0]
                yield Candle(
                    time=candle_time,
                    open=Decimal(c[1]),
                    high=Decimal(c[2]),
                    low=Decimal(c[3]),
                    close=Decimal(c[4]),
                    volume=Decimal(c[5]),
                )
                binance_start = candle_time + 1
            if len(content) < limit:
                return

    @asynccontextmanager
    async def connect_stream_candles(
        self, trading_pair: str, interval: int
    ) -> AsyncIterator[AsyncIterable[Candle]]:
        # Binance disconnects a websocket connection every 24h. Therefore, we reconnect every 12h.
        # Note that two streams will send events with matching evt_times.
        # This can be used to switch from one stream to another and avoiding the edge case where
        # we miss out on the very last update to a candle.

        async def inner(ws: AsyncIterable[Any]) -> AsyncIterable[Candle]:
            async for data in ws:
                c = data["k"]
                if c["x"]:  # Closed.
                    yield Candle(
                        time=c["t"],
                        open=Decimal(c["o"]),
                        high=Decimal(c["h"]),
                        low=Decimal(c["l"]),
                        close=Decimal(c["c"]),
                        volume=Decimal(c["v"]),
                    )

        async with self._connect_refreshing_stream(
            url=f"/ws/{_to_binance_ws_trading_pair(trading_pair)}@kline_{time.format_interval(interval)}",
            interval=12 * time.HOUR_MS,
            name="candles",
        ) as ws:
            yield inner(ws)

    # Trades.

    async def stream_historical_trades(
        self, trading_pair: str, start: int, end: int
    ) -> AsyncIterable[Trade]:
        # Aggregated trades. This means trades executed at the same time, same price and as part of
        # the same order will be aggregated by summing their size.
        batch_start = start
        payload: Dict[str, Any] = {
            "symbol": _to_binance_trading_pair(trading_pair),
        }
        while True:
            batch_end = batch_start + time.HOUR_MS
            payload["startTime"] = batch_start
            payload["endTime"] = min(batch_end, end) - 1  # Inclusive.

            trade_time = None

            content = await self._api_request(
                method="GET",
                url="/api/v3/aggTrades",
                data=payload,
            )
            for t in content:
                trade_time = t["T"]
                assert trade_time < end
                yield Trade(
                    id=t["a"],
                    time=trade_time,
                    price=Decimal(t["p"]),
                    size=Decimal(t["q"]),
                )
            batch_start = trade_time + 1 if trade_time is not None else batch_end
            if batch_start >= end:
                break

    @asynccontextmanager
    async def connect_stream_trades(
        self, trading_pair: str
    ) -> AsyncIterator[AsyncIterable[Trade]]:
        async def inner(ws: AsyncIterable[Any]) -> AsyncIterable[Trade]:
            async for data in ws:
                yield Trade(
                    id=data["a"],
                    time=data["T"],
                    price=Decimal(data["p"]),
                    size=Decimal(data["q"]),
                )

        async with self._connect_refreshing_stream(
            url=f"/ws/{_to_binance_ws_trading_pair(trading_pair)}@trade",
            interval=12 * time.HOUR_SEC,
            name="trades",
        ) as ws:
            yield inner(ws)

    # Savings.

    async def map_savings_products(self, asset: Optional[str] = None) -> Dict[str, SavingsProduct]:
        return {}

    async def purchase_savings_product(self, product_id: str, size: Decimal) -> None:
        _log.info(f"would have attempted to purchase {size} of {product_id}")

    async def redeem_savings_product(self, product_id: str, size: Decimal) -> None:
        _log.info(f"would have attempted to redeem {size} of {product_id}")

    # Common.

    async def _api_request(
        self,
        method: str,
        url: str,
        weight: int = 1,
        security: int = _SEC_NONE,
        data: Optional[Any] = None,
    ) -> Any:
        limiter_tasks = [
            self._raw_reqs_limiter.acquire(),
            self._reqs_per_min_limiter.acquire(weight),
        ]
        await asyncio.gather(*limiter_tasks)

        kwargs: Dict[str, Any] = {}

        if security == _SEC_USER_DATA:
            headers: Dict[str, str] = {}
            data = data or {}
            await self._authenticate_request(headers, data)
            if headers:
                kwargs["headers"] = headers
        if data:
            kwargs["params" if method == "GET" else "data"] = data

        try:
            response = await self._request(method=method, url=_BASE_API_URL + url, **kwargs)
        except BinanceClientException as exc:
            await self._intercept_client_exception(exc)
            raise

        return await response.json()

    async def _authenticate_request(self, headers: Dict[str, str], data: Dict[str, Any]) -> None:
        pass

    # We don't want to retry here because the caller of this method may need to adjust request
    # params on retry.
    async def _request(self, method: str, url: str, **kwargs: Any) -> ClientResponse:
        try:
            async with self._session.request(method=method, url=url, **kwargs) as response:
                # TODO: If status 50X (502 for example during exchange maintenance), we may
                # want to wait for a some kind of a successful health check before retrying.
                if response.status >= 500:
                    text_content = await response.text()
                    raise ConnectorException(
                        f"Server (5XX) error {response.status} {text_content}"
                    )
                if response.status >= 400:
                    try:
                        content = await response.json()
                        code = content.get("code")
                        message = content.get("msg")
                        if code is not None:
                            raise BinanceClientException(code, message, response.headers)
                    except aiohttp.client_exceptions.ContentTypeError:
                        pass
                    text_content = await response.text()
                    raise ConnectorException(
                        f"Client (4XX) error {response.status} {text_content}"
                    )
                return response
        except (
            aiohttp.ClientConnectionError,
            aiohttp.ClientPayloadError,
        ) as e:
            _log.warning(f"request exc: {e}")
            raise ConnectorException(str(e))

    @asynccontextmanager
    async def _connect_refreshing_stream(
        self,
        url: str,
        interval: float,
        name: str,
    ) -> AsyncIterator[AsyncIterable[Any]]:
        try:
            async with connect_refreshing_stream(
                self._session,
                url=_BASE_WS_URL + url,
                interval=interval,
                loads=partial(json.loads, use_decimal=True, parse_constant=Decimal),
                take_until=lambda old, new: old["E"] < new["E"],
                name=name,
                raise_on_disconnect=True,
            ) as stream:
                yield stream
        except (
            aiohttp.ClientConnectionError,
            aiohttp.ClientPayloadError,
            aiohttp.ClientResponseError,
            aiohttp.WebSocketError,
        ) as e:
            raise ConnectorException(str(e))

    async def _intercept_client_exception(
        self,
        exc: BinanceClientException,
    ) -> None:
        if exc.code == _ERR_TOO_MANY_REQUESTS:
            retry_after = exc.headers.get(istr("Retry-After"))
            if retry_after:
                _log.info(f"server provided retry-after {retry_after}; sleeping")
                await asyncio.sleep(float(retry_after))


class BinanceClientException(ConnectorException):
    """4XX errors specific to Binance."""

    def __init__(self, code: int, message: Optional[str], headers: CIMultiDictProxy) -> None:
        super().__init__(message)
        self.code = code
        self.headers = headers


def _to_binance_trading_pair(trading_pair: str) -> str:
    return trading_pair.replace("-", "")


def _to_binance_ws_trading_pair(trading_pair: str) -> str:
    return _to_binance_trading_pair(trading_pair).lower()



from typing import Generic, Optional, TypeVar

T = TypeVar("T")


class ChangedFilter(Generic[T]):
    """Pass a value only if it differs from previous value."""

    _value: Optional[T] = None
    _value_changed: bool = False

    @property
    def value(self) -> Optional[T]:
        return self._value if self._value_changed else None

    @property
    def prevailing_value(self) -> Optional[T]:
        return self._value

    def update(self, value: T) -> Optional[T]:
        self._value_changed = value != self._value
        self._value = value
        return self.value



import asyncio
import logging
import sqlite3
from collections import defaultdict
from contextlib import closing
from decimal import Decimal
from pathlib import Path
from typing import (
    Any,
    AsyncIterable,
    ContextManager,
    Dict,
    List,
    NamedTuple,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    get_type_hints,
)

from barbthroat import time
from barbthroat.itertools import generate_missing_spans, merge_adjacent_spans

_log = logging.getLogger(__name__)

# Version should be incremented every time a storage schema changes.
_VERSION = "v1"
_DATA_DIR = ".barbthroat"

T = TypeVar("T")

Primitive = Union[bool, int, float, Decimal, str]


def _serialize_decimal(d: Decimal) -> bytes:
    return str(d).encode("ascii")


def _deserialize_decimal(s: bytes) -> Decimal:
    return Decimal(s.decode("ascii"))


sqlite3.register_adapter(Decimal, _serialize_decimal)
sqlite3.register_converter("DECIMAL", _deserialize_decimal)

sqlite3.register_adapter(bool, int)
sqlite3.register_converter("BOOLEAN", lambda v: bool(int(v)))


class SQLite:
    def __init__(self) -> None:
        self._tables: Dict[Any, Set[str]] = defaultdict(set)

    async def stream_time_series_spans(
        self,
        shard: str,
        key: str,
        start: int = 0,
        end: int = time.MAX_TIME,
    ) -> AsyncIterable[Tuple[int, int]]:
        def inner() -> List[Tuple[int, int]]:
            _log.info(
                f"Streaming span(s) between {time.format_span(start, end)} from shard {shard} "
                f"{key}."
            )
            with self._connect(shard) as conn:
                span_key = f"{key}_{_SPAN_KEY}"
                self._ensure_table(conn, span_key, Span)
                return conn.execute(
                    f"SELECT * FROM {span_key} WHERE start < ? AND end > ? ORDER BY start",
                    [end, start],
                ).fetchall()

        rows = await asyncio.get_running_loop().run_in_executor(None, inner)
        for span_start, span_end in merge_adjacent_spans(rows):
            yield max(span_start, start), min(span_end, end)

    async def stream_time_series(
        self,
        shard: str,
        key: str,
        type_: Type[T],
        start: int = 0,
        end: int = time.MAX_TIME,
    ) -> AsyncIterable[T]:
        def inner() -> List[T]:
            _log.info(
                f"Streaming items between {time.format_span(start, end)} from shard {shard} {key}."
            )
            with self._connect(shard) as conn:
                self._ensure_table(conn, key, type_)
                return conn.execute(
                    f"SELECT * FROM {key} WHERE time >= ? AND time < ? ORDER BY time",
                    [start, end],
                ).fetchall()

        rows = await asyncio.get_running_loop().run_in_executor(None, inner)
        for row in rows:
            yield type_(*row)

    async def store_time_series_and_span(
        self, shard: str, key: str, items: List[Any], start: int, end: int
    ) -> None:
        # Even if items list is empty, we still want to store a span for the period!
        if len(items) > 0:
            type_ = type(items[0])
            if start > items[0].time:
                raise ValueError(
                    f"Span start {start} bigger than first item time {items[0].time}."
                )
            if end <= items[-1].time:
                raise ValueError(
                    f"Span end {end} smaller than or equal to last item time {items[-1].time}."
                )

        def inner() -> None:
            span_key = f"{key}_{_SPAN_KEY}"
            with self._connect(shard) as conn:
                self._ensure_table(conn, span_key, Span)
                if len(items) > 0:
                    self._ensure_table(conn, key, type_)

                c = conn.cursor()
                existing_spans = c.execute(
                    f"SELECT * FROM {span_key} WHERE start < ? AND end > ? ORDER BY start",
                    [end, start],
                ).fetchall()
                merged_existing_spans = merge_adjacent_spans(existing_spans)
                missing_spans = list(generate_missing_spans(start, end, merged_existing_spans))
                if len(missing_spans) == 0:
                    return
                missing_item_spans = (
                    [items]
                    if len(existing_spans) == 0
                    else [
                        [i for i in items if i.time >= s and i.time < e] for s, e in missing_spans
                    ]
                )
                for (mstart, mend), mitems in zip(missing_spans, missing_item_spans):
                    _log.info(
                        f"Inserting {len(mitems)} item(s) between "
                        f"{time.format_span(mstart, mend)} to shard {shard} {key}."
                    )
                    if len(mitems) > 0:
                        try:
                            c.executemany(
                                f"INSERT INTO {key} "
                                f'VALUES ({", ".join(["?"] * len(get_type_hints(type_)))})',
                                mitems,
                            )
                        except sqlite3.IntegrityError as err:
                            _log.error(f"{err} {shard} {key}")
                            raise
                    c.execute(f"INSERT INTO {span_key} VALUES (?, ?)", [mstart, mend])
                conn.commit()

        await asyncio.get_running_loop().run_in_executor(None, inner)

    def _connect(self, shard: str) -> ContextManager[sqlite3.Connection]:
        path = str(_home_path("data") / f"{_VERSION}_{shard}.db")
        _log.debug(f"Opening shard {path}.")
        return closing(sqlite3.connect(path, detect_types=sqlite3.PARSE_DECLTYPES))

    def _ensure_table(self, conn: sqlite3.Connection, name: str, type_: Type[Any]) -> None:
        tables = self._tables[conn]
        if name not in tables:
            c = conn.cursor()
            _create_table(c, type_, name)
            conn.commit()
            tables.add(name)


def _create_table(c: sqlite3.Cursor, type_: Type[Any], name: str) -> None:
    type_hints = get_type_hints(type_)
    col_types = [(k, _type_to_sql_type(v)) for k, v in type_hints.items()]

    # Create table.
    cols = []
    for col_name, col_type in col_types:
        cols.append(f"{col_name} {col_type} NOT NULL")
    c.execute(f'CREATE TABLE IF NOT EXISTS {name} ({", ".join(cols)})')

    # Add indices.
    meta_getter = getattr(type_, "meta", None)
    meta = meta_getter() if meta_getter else None
    if meta:
        for cname, ctype in meta.items():
            if ctype == "index":
                c.execute(f"CREATE INDEX IF NOT EXISTS {name}Index ON {name}({cname})")
            elif ctype == "unique":
                c.execute(
                    f"CREATE UNIQUE INDEX IF NOT EXISTS {name}UniqueIndex ON {name}({cname})"
                )
            else:
                raise NotImplementedError()


def _type_to_sql_type(type_: Type[Primitive]) -> str:
    if type_ is int:
        return "INTEGER"
    if type_ is float:
        return "REAL"
    if type_ is Decimal:
        return "DECIMAL"
    if type_ is str:
        return "TEXT"
    if type_ is bool:
        return "BOOLEAN"
    raise NotImplementedError(f"Missing conversion for type {type_}.")


def _home_path(*args: str) -> Path:
    path = Path(Path.home(), _DATA_DIR).joinpath(*args)
    path.mkdir(parents=True, exist_ok=True)
    return path


class Span(NamedTuple):
    start: int
    end: int

    @staticmethod
    def meta() -> Dict[str, str]:
        return {
            "start": "unique",
            "end": "unique",
        }


_SPAN_KEY = Span.__name__.lower()



from __future__ import annotations

import sqlite3
from collections import defaultdict
from contextlib import contextmanager
from threading import Lock
from types import TracebackType
from typing import Iterator, Optional

from .sqlite import SQLite


class Memory(SQLite):
    """In-memory data storage. Uses SQLite's memory mode for implementation."""

    def __init__(self) -> None:
        super().__init__()
        self._conns: dict[str, _ConnectionContext] = defaultdict(_ConnectionContext)

    async def __aenter__(self) -> Memory:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        for ctx in self._conns.values():
            assert ctx.connection
            ctx.connection.close()

    @contextmanager
    def _connect(self, shard: str) -> Iterator[sqlite3.Connection]:
        ctx = self._conns[shard]
        with ctx.lock:
            if not ctx.connection:
                conn = sqlite3.connect(
                    ":memory:",
                    detect_types=sqlite3.PARSE_DECLTYPES,
                    check_same_thread=False,
                )
                ctx.connection = conn
            yield ctx.connection


class _ConnectionContext:
    def __init__(self) -> None:
        self.connection: Optional[sqlite3.Connection] = None
        self.lock = Lock()



import asyncio
import logging
from decimal import Decimal
from typing import Dict, Optional

from barbthroat import hummingbot
from barbthroat.connectors import Connector
from barbthroat.models import SavingsProduct

_log = logging.getLogger(__name__)

_BALANCE_TIMEOUT = 10.0
_PRODUCT_TIMEOUT = 30 * 60.0

_SAVINGS_ASSET_PREFIX = "LD"


class SavingsCustodian:
    """
    On Binance, a flexible savings asset is indicated with an "LD"-prefix. It stands for "Lending
    Daily".
    """

    def __init__(
        self,
        connectors: Dict[str, Connector],
        hummingbot_connectors: Dict[str, hummingbot.Connector],
    ) -> None:
        self._connectors = connectors
        self._hummingbot_connectors = hummingbot_connectors

    def to_savings_asset(self, asset: str) -> str:
        return (
            asset if asset.startswith(_SAVINGS_ASSET_PREFIX) else f"{_SAVINGS_ASSET_PREFIX}{asset}"
        )

    def from_savings_asset(self, asset: str) -> str:
        return asset[2:] if asset.startswith(_SAVINGS_ASSET_PREFIX) else asset

    async def acquire(self, connector_name: str, asset: str, amount: Decimal) -> None:
        _log.info(f"Request to redeem {amount} worth of {asset} flexible savings product.")

        product = await asyncio.wait_for(
            self._wait_for_product_status_purchasing(connector_name, asset),
            timeout=_PRODUCT_TIMEOUT,
        )
        if product is None:
            _log.info(f"{asset} savings product not available; skipping.")
            return

        savings_asset = self.to_savings_asset(asset)
        savings_amount = self._hummingbot_connectors[connector_name].get_available_balance(
            savings_asset
        )
        if savings_amount == 0:
            _log.info("Nothing to redeem; savings balance 0.")
            return

        await self._connectors[connector_name].redeem_savings_product(
            product.product_id, savings_amount
        )
        await asyncio.wait_for(
            self._wait_for_wallet_updated_with(connector_name, asset, savings_amount),
            timeout=_BALANCE_TIMEOUT,
        )

        _log.info(f"Redeemed {savings_amount} worth of {asset} flexible savings product.")

    async def release(self, connector_name: str, asset: str, amount: Decimal) -> None:
        _log.info(f"Request to purchase {amount} worth of {asset} flexible savings product.")

        product = await asyncio.wait_for(
            self._wait_for_product_status_purchasing(connector_name, asset),
            timeout=_PRODUCT_TIMEOUT,
        )
        if product is None:
            _log.info(f"{asset} savings product not available; skipping.")
            return

        savings_amount = amount

        global_available_product = product.limit - product.purchased_amount
        if amount > global_available_product:
            _log.info(f"Only {global_available_product} available globally.")
            savings_amount = global_available_product

        if amount > product.limit_per_user:
            _log.info(f"Only {product.limit_per_user} available per user.")
            savings_amount = product.limit_per_user

        if savings_amount < product.min_purchase_amount:
            _log.info(
                f"{savings_amount} less than minimum purchase amount "
                f"{product.min_purchase_amount}; skipping."
            )
            return

        await self._connectors[connector_name].purchase_savings_product(
            product.product_id, savings_amount
        )
        await asyncio.wait_for(
            self._wait_for_wallet_updated_with(
                connector_name, self.to_savings_asset(asset), savings_amount
            ),
            timeout=_BALANCE_TIMEOUT,
        )

        _log.info(f"Purchased {savings_amount} worth of {asset} flexible savings product.")

    async def _wait_for_wallet_updated_with(
        self, connector_name: str, asset: str, amount: Decimal
    ) -> None:
        while True:
            # TODO: Ideally we listened to websocket updates instead of polling the wallet.
            available_amount = self._hummingbot_connectors[connector_name].get_available_balance(
                asset
            )
            if available_amount >= amount:
                return
            await asyncio.sleep(100)

    async def _wait_for_product_status_purchasing(
        self, connector_name: str, asset: str
    ) -> Optional[SavingsProduct]:
        # A product can be in status "PURCHASING" or "PREHEATING". "PURCHASING" is when the product
        # is available. "PREHEATING" means the product is being processed. This happens usually at
        # 23:50 - 00:10 UTC.
        # https://dev.binance.vision/t/failure-to-fast-redeem-a-flexible-savings-product-right-after-midnight-00-00-utc/5785
        while True:
            products = await self._connectors[connector_name].map_savings_products(asset=asset)

            product = products.get(asset)
            if product is None:
                return None

            if product.status == "PREHEATING":
                _log.info(
                    f"{asset} savings product is preheating; waiting a minute before retrying."
                )
                await asyncio.sleep(60.0)
            elif product.status == "PURCHASING":
                return product
            else:
                raise Exception(f"Unknown {asset} savings product status {product.status}.")



from dataclasses import dataclass
from decimal import Decimal
from typing import Dict, Literal, NamedTuple


class Candle(NamedTuple):
    # Interval start time.
    time: int
    open: Decimal
    high: Decimal
    low: Decimal
    close: Decimal
    volume: Decimal

    @staticmethod
    def meta() -> Dict[str, str]:
        return {
            "time": "unique",
        }


class Trade(NamedTuple):
    # Aggregate trade id.
    id: int = 0
    # Can have multiple trades at same time.
    time: int = 0
    price: Decimal = Decimal("0.0")
    size: Decimal = Decimal("0.0")

    @staticmethod
    def meta() -> Dict[str, str]:
        return {
            "time": "index",
        }


@dataclass(frozen=True)
class SavingsProduct:
    product_id: str
    status: Literal["PREHEATING", "PURCHASING"]
    asset: str
    can_purchase: bool
    can_redeem: bool
    purchased_amount: Decimal
    min_purchase_amount: Decimal
    limit: Decimal
    limit_per_user: Decimal

    @property
    def max_purchase_amount_for_user(self) -> Decimal:
        return min(self.limit - self.purchased_amount, self.limit_per_user)


class ConnectorException(Exception):
    """Covers errors both on the client (4XX) and server (5XX) side. Operation can be retried."""

    pass



from typing import Iterable, Tuple


def merge_adjacent_spans(spans: Iterable[Tuple[int, int]]) -> Iterable[Tuple[int, int]]:
    merged_start, merged_end = None, None

    for start, end in spans:
        if merged_start is None:
            merged_start, merged_end = start, end
        elif merged_end == start:
            merged_end = end
        else:
            yield merged_start, merged_end
            merged_start, merged_end = start, end

    if merged_start is not None and merged_end is not None:
        yield merged_start, merged_end


def generate_missing_spans(
    start: int, end: int, existing_spans: Iterable[Tuple[int, int]]
) -> Iterable[Tuple[int, int]]:
    # Initially assume entire span missing.
    missing_start, missing_end = start, end

    # Spans are ordered by start_date. Spans do not overlap with each other.
    for existing_start, existing_end in existing_spans:
        if existing_start > missing_start:
            yield missing_start, existing_start
        missing_start = existing_end

    if missing_start < missing_end:
        yield missing_start, missing_end



import asyncio
import logging
import sys
from collections import deque
from contextlib import AsyncExitStack
from decimal import Decimal
from typing import Any, AsyncGenerator, AsyncIterable, Deque, Dict, List, Optional

from asyncstdlib import list as list_async
from tenacity import AsyncRetrying, before_sleep_log, retry_if_exception_type

from barbthroat import time
from barbthroat.connectors import Connector
from barbthroat.itertools import generate_missing_spans
from barbthroat.models import Candle, ConnectorException, Trade
from barbthroat.storages import Storage
from barbthroat.tenacity import stop_after_attempt_with_reset, wait_none_then_exponential

_log = logging.getLogger(__name__)

_CANDLE_KEY = Candle.__name__.lower()
_TRADE_KEY = Trade.__name__.lower()

_STORAGE_BATCH_SIZE = 1000


class Chandler:
    """
    A component responsible for providing candlestick info. Features:

    - Seamlessly merge together historical and future candles.
    - Construct candles from trades.
    - Store candles locally for quicker retrieval and to minimize communication with an exchange.
    """

    def __init__(
        self,
        storage: Storage,
        connectors: Dict[str, Connector],
    ) -> None:
        self._storage = storage
        self._connectors = connectors

    async def stream_candles(
        self,
        connector_name: str,
        trading_pair: str,
        interval: int,
        start: int,
        end: int = time.MAX_TIME,
    ) -> AsyncIterable[Candle]:
        """
        Tries to stream candles for the specified range from local storage. If candles don't exist,
        streams them from an exchange and stores to local storage.
        """
        start = time.floor_timestamp(start, interval)
        end = time.floor_timestamp(end, interval)

        if end <= start:
            return

        shard = _key(connector_name, trading_pair, interval)
        candle_msg = f"{connector_name} {trading_pair} {time.format_interval(interval)} candle(s)"

        _log.info(f"Checking for existing {candle_msg} in local storage.")
        existing_spans = await list_async(
            self._storage.stream_time_series_spans(
                shard=shard,
                key=_CANDLE_KEY,
                start=start,
                end=end,
            )
        )
        missing_spans = list(generate_missing_spans(start, end, existing_spans))

        spans = [(a, b, True) for a, b in existing_spans] + [
            (a, b, False) for a, b in missing_spans
        ]
        spans.sort(key=lambda s: s[0])

        last_candle: Optional[Candle] = None
        for span_start, span_end, exist_locally in spans:
            period_msg = f"{time.format_span(span_start, span_end)}"
            if exist_locally:
                _log.info(f"Local {candle_msg} exist between {period_msg}.")
                stream = self._storage.stream_time_series(
                    shard=shard,
                    key=_CANDLE_KEY,
                    type_=Candle,
                    start=span_start,
                    end=span_end,
                )
            else:
                _log.info(f"Missing {candle_msg} between {period_msg}.")
                stream = self._stream_and_store_exchange_candles(
                    connector_name=connector_name,
                    trading_pair=trading_pair,
                    interval=interval,
                    start=span_start,
                    end=span_end,
                )
            try:
                async for candle in stream:
                    if last_candle:
                        time_diff = candle.time - last_candle.time
                        if time_diff >= interval * 2:
                            num_missed = time_diff // interval - 1
                            _log.warning(
                                f"Missed {num_missed} {candle_msg}; last closed candle "
                                f"{last_candle}. current candle {candle}."
                            )
                    else:
                        num_missed = (candle.time - start) // interval
                        if num_missed > 0:
                            _log.warning(
                                f"Missed {num_missed} {candle_msg} from the start "
                                f"{time.format_timestamp(start)}; current candle {candle}."
                            )

                    yield candle
                    last_candle = candle
            finally:
                if isinstance(stream, AsyncGenerator):
                    await stream.aclose()

        if last_candle:
            time_diff = end - last_candle.time
            if time_diff >= interval * 2:
                num_missed = time_diff // interval - 1
                _log.warning(
                    f"Missed {num_missed} {candle_msg} from the end {time.format_timestamp(end)}; "
                    f"current candle {last_candle}."
                )
        else:
            _log.warning(f"Missed all {candle_msg} between {time.format_span(start, end)}.")

    async def _stream_and_store_exchange_candles(
        self,
        connector_name: str,
        trading_pair: str,
        interval: int,
        start: int,
        end: int,
    ) -> AsyncGenerator[Candle, None]:
        shard = _key(connector_name, trading_pair, interval)
        # Note that we need to use a context manager based retrying because retry decorators do not
        # work with async generator functions.
        async for attempt in AsyncRetrying(
            stop=stop_after_attempt_with_reset(8, 300),
            wait=wait_none_then_exponential(),
            retry=retry_if_exception_type(ConnectorException),
            before_sleep=before_sleep_log(_log, logging.WARNING),
        ):
            with attempt:
                # We use a swap batch in order to swap the batch right before storing. With a
                # single batch, it may happen that our program gets cancelled at an `await` point
                # before we're able to clear the batch. This can cause same data to be stored
                # twice, raising an integrity error.
                batch = []
                swap_batch: List[Candle] = []
                current = time.floor_timestamp(time.now(), interval)

                try:
                    stream = self._stream_exchange_candles(
                        connector_name=connector_name,
                        trading_pair=trading_pair,
                        interval=interval,
                        start=start,
                        end=end,
                        current=current,
                    )
                    try:
                        async for candle in stream:
                            batch.append(candle)
                            if len(batch) == _STORAGE_BATCH_SIZE:
                                del swap_batch[:]
                                batch_start = start
                                batch_end = batch[-1].time + interval
                                start = batch_end
                                swap_batch, batch = batch, swap_batch
                                await self._storage.store_time_series_and_span(
                                    shard=shard,
                                    key=_CANDLE_KEY,
                                    items=swap_batch,
                                    start=batch_start,
                                    end=batch_end,
                                )
                            yield candle
                    finally:
                        await stream.aclose()
                except (asyncio.CancelledError, ConnectorException):
                    if len(batch) > 0:
                        batch_start = start
                        batch_end = batch[-1].time + interval
                        start = batch_end
                        await self._storage.store_time_series_and_span(
                            shard=shard,
                            key=_CANDLE_KEY,
                            items=batch,
                            start=batch_start,
                            end=batch_end,
                        )
                    raise
                else:
                    current = time.floor_timestamp(time.now(), interval)
                    await self._storage.store_time_series_and_span(
                        shard=shard,
                        key=_CANDLE_KEY,
                        items=batch,
                        start=start,
                        end=min(current, end),
                    )

    async def _stream_exchange_candles(
        self,
        connector_name: str,
        trading_pair: str,
        interval: int,
        start: int,
        end: int,
        current: int,
    ) -> AsyncGenerator[Candle, None]:
        exchange_instance = self._connectors[connector_name]
        intervals = exchange_instance.list_candle_intervals()
        is_candle_interval_supported = interval in intervals

        async def inner(stream: Optional[AsyncIterable[Candle]]) -> AsyncGenerator[Candle, None]:
            if start < current:  # Historical.
                historical_end = min(end, current)
                if (
                    exchange_instance.can_stream_historical_candles
                    and is_candle_interval_supported
                ):
                    historical_stream = exchange_instance.stream_historical_candles(
                        trading_pair=trading_pair,
                        interval=interval,
                        start=start,
                        end=historical_end,
                    )
                else:
                    historical_stream = self._stream_construct_candles(
                        connector_name, trading_pair, interval, start, historical_end
                    )
                try:
                    async for candle in historical_stream:
                        yield candle
                finally:
                    if isinstance(historical_stream, AsyncGenerator):
                        await historical_stream.aclose()
            if stream:  # Future.
                try:
                    async for candle in stream:
                        # If we start the websocket connection while candle is closing, we can also
                        # receive the same candle from here that we already got from historical.
                        # Ignore such candles.
                        if candle.time < current:
                            continue

                        if candle.time >= end:
                            break

                        yield candle

                        if candle.time == end - interval:
                            break
                finally:
                    if isinstance(stream, AsyncGenerator):
                        await stream.aclose()

        async with AsyncExitStack() as stack:
            stream = None
            if end > current:
                if exchange_instance.can_stream_candles and is_candle_interval_supported:
                    stream = await stack.enter_async_context(
                        exchange_instance.connect_stream_candles(trading_pair, interval)
                    )
                else:
                    stream = self._stream_construct_candles(
                        connector_name, trading_pair, interval, current, end
                    )

            last_candle_time = -1
            outer_stream = inner(stream)
            try:
                async for candle in outer_stream:
                    if interval < time.WEEK_MS and (candle.time % interval) != 0:
                        adjusted_time = time.floor_timestamp(candle.time, interval)
                        _log.warning(
                            f"Candle with bad time {candle} for interval "
                            f"{time.format_interval(interval)}; trying to adjust back in time to "
                            f"{time.format_timestamp(adjusted_time)} or skip if volume zero."
                        )
                        if last_candle_time == adjusted_time:
                            if candle.volume > 0:
                                raise RuntimeError(
                                    f"Received {trading_pair} {time.format_interval(interval)} "
                                    f"candle {candle} with a time that does not fall into the "
                                    "interval. Cannot adjust back in time because time coincides "
                                    "with last candle time "
                                    f"{time.format_timestamp(last_candle_time)}. Cannot skip "
                                    "because volume not zero."
                                )
                            else:
                                continue
                        candle = Candle(
                            time=adjusted_time,
                            open=candle.open,
                            high=candle.high,
                            low=candle.low,
                            close=candle.close,
                            volume=candle.volume,
                        )

                    yield candle
                    last_candle_time = candle.time
            finally:
                await outer_stream.aclose()

    async def _stream_construct_candles(
        self,
        connector_name: str,
        trading_pair: str,
        interval: int,
        start: int,
        end: int,
    ) -> AsyncGenerator[Candle, None]:
        _log.info(f"Constructing {connector_name} {trading_pair} {interval} candles from trades.")

        current = start
        next_ = current + interval
        open_ = Decimal("0.0")
        high = Decimal("0.0")
        low = Decimal(f"{sys.maxsize}.0")
        close = Decimal("0.0")
        volume = Decimal("0.0")
        is_first = True
        async for trade in self.stream_trades(
            connector_name=connector_name,
            trading_pair=trading_pair,
            start=start,
            end=end,
        ):
            if trade.time >= next_:
                assert not is_first
                yield Candle(
                    time=current,
                    open=open_,
                    high=high,
                    low=low,
                    close=close,
                    volume=volume,
                )
                current = next_
                next_ = current + interval
                open_ = Decimal("0.0")
                high = Decimal("0.0")
                low = Decimal(f"{sys.maxsize}.0")
                close = Decimal("0.0")
                volume = Decimal("0.0")
                is_first = True

            if is_first:
                open_ = trade.price
                is_first = False
            high = max(high, trade.price)
            low = min(low, trade.price)
            close = trade.price
            volume += trade.size

        if not is_first:
            yield Candle(
                time=current,
                open=open_,
                high=high,
                low=low,
                close=close,
                volume=volume,
            )

    # Trades

    async def stream_trades(
        self,
        connector_name: str,
        trading_pair: str,
        start: int,
        end: int,
    ) -> AsyncIterable[Trade]:
        """
        Tries to stream trades for the specified range from local storage. If trades don't exist,
        streams them from an exchange and stores to local storage.
        """
        shard = _key(connector_name, trading_pair)
        trade_msg = f"{connector_name} {trading_pair} trades"

        _log.info(f"Checking for existing {trade_msg} in local storage.")
        existing_spans = await list_async(
            self._storage.stream_time_series_spans(
                shard=shard,
                key=_TRADE_KEY,
                start=start,
                end=end,
            )
        )
        missing_spans = list(generate_missing_spans(start, end, existing_spans))

        spans = [(a, b, True) for a, b in existing_spans] + [
            (a, b, False) for a, b in missing_spans
        ]
        spans.sort(key=lambda s: s[0])

        for span_start, span_end, exist_locally in spans:
            period_msg = f"{time.format_span(span_start, span_end)}"
            if exist_locally:
                _log.info(f"Local {trade_msg} exist between {period_msg}.")
                stream = self._storage.stream_time_series(
                    shard=shard,
                    key=_TRADE_KEY,
                    type_=Trade,
                    start=span_start,
                    end=span_end,
                )
            else:
                _log.info(f"Missing {trade_msg} between {period_msg}.")
                stream = self._stream_and_store_exchange_trades(
                    connector_name, trading_pair, span_start, span_end
                )
            async for trade in stream:
                yield trade

    async def _stream_and_store_exchange_trades(
        self,
        connector_name: str,
        trading_pair: str,
        start: int,
        end: int,
    ) -> AsyncIterable[Trade]:
        shard = _key(connector_name, trading_pair)
        async for attempt in AsyncRetrying(
            stop=stop_after_attempt_with_reset(8, 300),
            wait=wait_none_then_exponential(),
            retry=retry_if_exception_type(ConnectorException),
            before_sleep=before_sleep_log(_log, logging.WARNING),
        ):
            with attempt:
                # We use a swap batch in order to swap the batch right before storing. With a
                # single batch, it may happen that our program gets cancelled at an `await` point
                # before we're able to clear the batch. This can cause same data to be stored
                # twice, raising an integrity error. We also use swap to store trades from previous
                # batch in case we get multiple trades with a same time at the edge of the batch.
                batch = []
                swap_batch: List[Trade] = []
                current = time.now()

                try:
                    async for trade in self._stream_exchange_trades(
                        connector_name=connector_name,
                        trading_pair=trading_pair,
                        start=start,
                        end=end,
                        current=current,
                    ):
                        batch.append(trade)
                        # We go over limit with +1 because we never take the last trade of the
                        # batch because multiple trades can happen at the same time. We need our
                        # time span to be correct.
                        if len(batch) == _STORAGE_BATCH_SIZE + 1:
                            del swap_batch[:]

                            last = batch[-1]
                            for i in range(len(batch) - 1, -1, -1):
                                if batch[i].time != last.time:
                                    break
                                # Note that we are inserting in front.
                                swap_batch.insert(0, batch[i])
                                del batch[i]

                            batch_start = start
                            batch_end = batch[-1].time + 1
                            swap_batch, batch = batch, swap_batch
                            start = batch_end
                            await self._storage.store_time_series_and_span(
                                shard=shard,
                                key=_TRADE_KEY,
                                items=swap_batch,
                                start=batch_start,
                                end=batch_end,
                            )
                        yield trade
                except (asyncio.CancelledError, ConnectorException):
                    if len(batch) > 0:
                        batch_start = start
                        batch_end = batch[-1].time + 1
                        start = batch_end
                        await self._storage.store_time_series_and_span(
                            shard=shard,
                            key=_TRADE_KEY,
                            items=batch,
                            start=batch_start,
                            end=batch_end,
                        )
                    raise
                else:
                    current = time.now()
                    await self._storage.store_time_series_and_span(
                        shard=shard,
                        key=_TRADE_KEY,
                        items=batch,
                        start=start,
                        end=min(current, end),
                    )

    async def _stream_exchange_trades(
        self,
        connector_name: str,
        trading_pair: str,
        start: int,
        end: int,
        current: int,
    ) -> AsyncIterable[Trade]:
        connector = self._connectors[connector_name]

        async def inner(stream: Optional[AsyncIterable[Trade]]) -> AsyncIterable[Trade]:
            last_trade_ids: Deque[int] = deque(maxlen=20)
            if start < current:  # Historical.
                async for trade in connector.stream_historical_trades(
                    trading_pair, start, min(end, current)
                ):
                    if trade.id > 0:
                        last_trade_ids.append(trade.id)
                    yield trade
            if stream:  # Future.
                skipping_existing = True
                async for trade in stream:
                    # TODO: Can we improve? We may potentially wait for a long time before a trade
                    # past the end time occurs.
                    if trade.time >= end:
                        break

                    # Skip if trade was already retrieved from historical. If we start the
                    # websocket connection during a trade, we can also receive the same trade from
                    # here that we already got from historical.
                    if skipping_existing and (
                        trade.id > 0 and trade.id in last_trade_ids or trade.time < current
                    ):
                        continue
                    else:
                        skipping_existing = False

                    yield trade

        if end > current:
            async with connector.connect_stream_trades(trading_pair) as stream:
                async for trade in inner(stream):
                    yield trade
        else:
            async for trade in inner(None):
                yield trade


def _key(*items: Any) -> str:
    return "_".join(map(str.lower, map(str, items)))



from typing import Union

from .savings import SavingsCustodian
from .spot import SpotCustodian

Custodian = Union[SavingsCustodian, SpotCustodian]

__all__ = [
    "SavingsCustodian",
    "SpotCustodian",
]



from .changed_filter import ChangedFilter
from .four_week_rule import FourWeekRule
from .simple_moving_average import SimpleMovingAverage

__all__ = [
    "ChangedFilter",
    "FourWeekRule",
    "SimpleMovingAverage",
]



from typing import Union

from .binance import BinanceConnector
from .binance_paper_trade import BinancePaperTradeConnector

Connector = Union[BinanceConnector, BinancePaperTradeConnector]

__all__ = [
    "BinanceConnector",
    "BinancePaperTradeConnector",
]



import time
from datetime import datetime, timezone

from barbthroat.math import floor_multiple, floor_multiple_offset

SEC_MS = 1000
MIN_MS = 60_000
HOUR_MS = 3_600_000
DAY_MS = 86_400_000
WEEK_MS = 604_800_000
MONTH_MS = 2_629_746_000
YEAR_MS = 31_556_952_000

MIN_SEC = 60
HOUR_SEC = 3600
DAY_SEC = 86_400
WEEK_SEC = 604_800
MONTH_SEC = 2_629_746
YEAR_SEC = 31_556_952

# 2065-01-24 05:20
MAX_TIME = 3_000_000_000_000

_WEEK_OFFSET_MS = 345_600_000


def now() -> int:
    """Returns current time since EPOCH in milliseconds."""
    return int(round(time.time() * 1000.0))


def format_span(start: int, end: int) -> str:
    return f"{_to_datetime_utc(start)} - " f"{_to_datetime_utc(end)}"


def _to_datetime_utc(ms: int) -> datetime:
    return datetime.utcfromtimestamp(ms / 1000.0).replace(tzinfo=timezone.utc)


def _from_datetime_utc(dt: datetime) -> int:
    assert dt.tzinfo == timezone.utc
    return int(round(dt.timestamp() * 1000.0))


def format_interval(interval: int) -> str:
    result = ""
    remainder = interval
    for letter, factor in _INTERVAL_FACTORS.items():
        quotient, remainder = divmod(remainder, factor)
        if quotient > 0:
            result += f"{quotient}{letter}"
        if remainder == 0:
            break
    return result if result else "0ms"


def format_timestamp(timestamp: int) -> str:
    return _to_datetime_utc(timestamp).isoformat()


def floor_timestamp(timestamp: int, interval: int) -> int:
    if interval < WEEK_MS:
        return floor_multiple(timestamp, interval)
    if interval == WEEK_MS:
        return floor_multiple_offset(timestamp, interval, _WEEK_OFFSET_MS)
    if interval == MONTH_MS:
        dt = _to_datetime_utc(timestamp)
        return _from_datetime_utc(dt.replace(day=1, hour=0, minute=0, second=0, microsecond=0))
    raise NotImplementedError()


# Is assumed to be ordered by values descending.
_INTERVAL_FACTORS = {
    "y": YEAR_MS,
    "M": MONTH_MS,
    "w": WEEK_MS,
    "d": DAY_MS,
    "h": HOUR_MS,
    "m": MIN_MS,
    "s": SEC_MS,
    "ms": 1,
}



from __future__ import annotations

import asyncio
import hashlib
import hmac
import logging
import urllib.parse
from contextlib import suppress
from decimal import Decimal
from types import TracebackType
from typing import Any, Dict, Optional

from tenacity import (
    before_sleep_log,
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

from barbthroat import time
from barbthroat.models import ConnectorException, SavingsProduct

from .binance_paper_trade import _SEC_USER_DATA, BinanceClientException, BinancePaperTradeConnector

_log = logging.getLogger(__name__)

_ERR_INVALID_TIMESTAMP = -1021


class BinanceConnector(BinancePaperTradeConnector):
    def __init__(self, api_key: str, secret_key: str) -> None:
        super().__init__()
        self._api_key = api_key
        self._secret_key_bytes = secret_key.encode("utf-8")

        self._clock = Clock(self)

    async def __aenter__(self) -> BinanceConnector:
        await super().__aenter__()
        await self._clock.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        await self._clock.__aexit__(exc_type, exc, tb)
        await super().__aexit__(exc_type, exc, tb)

    # Savings.

    async def map_savings_products(self, asset: Optional[str] = None) -> Dict[str, SavingsProduct]:
        # TODO: This endpoint is paginated but only fetches the first page.
        data: Dict[str, Any] = {
            "size": 100,
        }
        if asset is not None:
            data["asset"] = asset
        content = await self._api_request(
            method="GET",
            url="/sapi/v1/lending/daily/product/list",
            data=data,
            security=_SEC_USER_DATA,
        )
        result = {}
        for product in content:
            product_asset = product["asset"]
            result[product_asset] = SavingsProduct(
                product_id=product["productId"],
                status=product["status"],
                asset=product_asset,
                can_purchase=product["canPurchase"],
                can_redeem=product["canRedeem"],
                purchased_amount=Decimal(product["purchasedAmount"]),
                min_purchase_amount=Decimal(product["minPurchaseAmount"]),
                limit=Decimal(product["upLimit"]),
                limit_per_user=Decimal(product["upLimitPerUser"]),
            )
        return result

    async def purchase_savings_product(self, product_id: str, size: Decimal) -> None:
        await self._api_request(
            method="POST",
            url="/sapi/v1/lending/daily/purchase",
            data={
                "productId": product_id,
                "amount": _to_decimal(size),
            },
            security=_SEC_USER_DATA,
        )

    async def redeem_savings_product(self, product_id: str, size: Decimal) -> None:
        await self._api_request(
            method="POST",
            url="/sapi/v1/lending/daily/redeem",
            data={
                "productId": product_id,
                "amount": _to_decimal(size),
                "type": "FAST",  # "FAST" | "NORMAL"
            },
            security=_SEC_USER_DATA,
        )

    # Common.

    async def _authenticate_request(self, headers: Dict[str, str], data: Dict[str, Any]) -> None:
        await self._clock.wait()

        headers["X-MBX-APIKEY"] = self._api_key

        data["timestamp"] = str(time.now() + self._clock.time_diff)

        query_str_bytes = urllib.parse.urlencode(data).encode("utf-8")
        signature = hmac.new(self._secret_key_bytes, query_str_bytes, hashlib.sha256)
        data["signature"] = signature.hexdigest()

    async def _intercept_client_exception(
        self,
        exc: BinanceClientException,
    ) -> None:
        await super()._intercept_client_exception(exc)
        if exc.code == _ERR_INVALID_TIMESTAMP:
            _log.warning("received invalid timestamp; syncing clock before exc")
            self._clock.clear()


class Clock:
    def __init__(self, binance: BinancePaperTradeConnector) -> None:
        self.time_diff = 0
        self._binance = binance
        self._synced = asyncio.Event()
        self._periodic_sync_task: Optional[asyncio.Task[None]] = None
        self._reset_periodic_sync: asyncio.Event = asyncio.Event()

    async def __aenter__(self) -> Clock:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        if self._periodic_sync_task:
            self._periodic_sync_task.cancel()
            with suppress(asyncio.CancelledError):
                await self._periodic_sync_task

    async def wait(self) -> None:
        if not self._periodic_sync_task:
            self._periodic_sync_task = asyncio.create_task(self._periodic_sync())

        await self._synced.wait()

    def clear(self) -> None:
        self._synced.clear()
        if self._periodic_sync_task:
            self._reset_periodic_sync.set()

    async def _periodic_sync(self) -> None:
        while True:
            await self._sync_clock()
            # 6 hours.
            sleep_task = asyncio.create_task(asyncio.sleep(time.HOUR_SEC * 6))
            reset_periodic_sync_task = asyncio.create_task(self._reset_periodic_sync.wait())
            try:
                await asyncio.wait(  # type:ignore
                    [sleep_task, reset_periodic_sync_task],
                    return_when=asyncio.FIRST_COMPLETED,
                )
            finally:
                if not sleep_task.done():
                    sleep_task.cancel()
                    with suppress(asyncio.CancelledError):
                        await sleep_task
                if not reset_periodic_sync_task.done():
                    reset_periodic_sync_task.cancel()
                    with suppress(asyncio.CancelledError):
                        await reset_periodic_sync_task
                self._reset_periodic_sync.clear()

    @retry(
        stop=stop_after_attempt(10),
        wait=wait_exponential(),
        retry=retry_if_exception_type(ConnectorException),
        before_sleep=before_sleep_log(_log, logging.WARNING),
    )
    async def _sync_clock(self) -> None:
        # https://github.com/binance-exchange/binance-official-api-docs/blob/master/rest-api.md#check-server-time
        _log.info("syncing clock with Binance")
        before = time.now()
        content = await self._binance._api_request(
            method="GET",
            url="/api/v3/time",
        )
        server_time = content["serverTime"]
        after = time.now()
        # Assume response time is same as request time.
        delay = (after - before) // 2
        local_time = before + delay
        # Adjustment required converting from local time to server time.
        self.time_diff = server_time - local_time
        _log.info(f"found {self.time_diff}ms time difference")
        self._synced.set()


def _to_decimal(value: Decimal) -> str:
    # Converts from scientific notation.
    # 6.4E-7 -> 0.0000_0064
    return f"{value:f}"



from time import time
from typing import Callable

from tenacity import RetryCallState
from tenacity.stop import stop_base
from tenacity.wait import wait_base, wait_chain, wait_exponential, wait_none


class stop_after_attempt_with_reset(stop_base):
    """
    Stop when the previous attempt >= max_attempt. Reset attempt count after time_to_reset seconds.
    """

    def __init__(
        self,
        max_attempt_number: int,
        time_to_reset: float,
        get_time: Callable[[], float] = time,
    ) -> None:
        self._max_attempt_number = max_attempt_number
        self._time_to_reset = time_to_reset
        self._get_time = get_time
        self._last_attempt_at = 0.0
        self._attempt_offset = 0

    def __call__(self, retry_state: RetryCallState) -> bool:
        now = self._get_time()
        if now - self._last_attempt_at >= self._time_to_reset:
            self._attempt_offset = retry_state.attempt_number - 1
        self._last_attempt_at = now
        return retry_state.attempt_number - self._attempt_offset >= self._max_attempt_number


def wait_none_then_exponential() -> wait_base:
    return wait_chain(wait_none(), wait_exponential())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['barbthroat',
 'barbthroat.connectors',
 'barbthroat.custodians',
 'barbthroat.indicators',
 'barbthroat.storages']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.3,<4.0.0',
 'aiolimiter>=1.0.0,<2.0.0',
 'asyncstdlib>=3.10.5,<4.0.0',
 'simplejson>=3.18.0,<4.0.0',
 'tenacity>=8.1.0,<9.0.0']

setup_kwargs = {
    'name': 'barbthroat',
    'version': '0.1.2',
    'description': 'A library to help write directional strategies for Hummingbot.',
    'long_description': '# 🐦barbthroat\n\n[![PyPI Downloads](https://img.shields.io/pypi/v/barbthroat.svg?label=PyPI)](\nhttps://pypi.org/project/barbthroat/)\n\nA library to help write directional strategies for [Hummingbot](https://hummingbot.org/).\n',
    'author': 'discosultan',
    'author_email': 'jaanusvarus@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from collections import deque
from decimal import Decimal
from typing import Deque


class SimpleMovingAverage:
    """
    Calculated by taking the arithmetic mean of a given set of values over a specified period.
    """

    _prices: Deque[Decimal]
    _value: Decimal = Decimal("0.0")

    @property
    def value(self) -> Decimal:
        return self._value

    def __init__(self, period: int) -> None:
        if period < 1:
            raise ValueError(f"Period must be longer than zero; got {period}.")

        self._prices = deque(maxlen=period)

    def update(self, price: Decimal) -> Decimal:
        self._prices.append(price)
        self._value = sum(self._prices, Decimal("0.0")) / len(self._prices)
        return self._value



from decimal import Decimal


class SpotCustodian:
    def to_savings_asset(self, asset: str) -> str:
        return asset

    def from_savings_asset(self, asset: str) -> str:
        return asset

    async def acquire(self, connector_name: str, asset: str, amount: Decimal) -> None:
        pass

    async def release(self, connector_name: str, asset: str, amount: Decimal) -> None:
        pass



from __future__ import annotations

import asyncio
import logging
import warnings
from contextlib import asynccontextmanager, suppress
from decimal import Decimal
from functools import partial
from itertools import cycle
from random import randint
from types import TracebackType
from typing import (
    Any,
    AsyncContextManager,
    AsyncIterable,
    AsyncIterator,
    Callable,
    Iterator,
    List,
    Optional,
    TypeVar,
)

import aiohttp
import simplejson as json
from asyncstdlib import chain as chain_async
from multidict import CIMultiDictProxy

_log = logging.getLogger(__name__)

T = TypeVar("T")


class ClientResponse:
    def __init__(self, response: aiohttp.ClientResponse) -> None:
        self._response = response

    @property
    def status(self) -> int:
        return self._response.status

    @property
    def headers(self) -> CIMultiDictProxy:
        return self._response.headers

    async def text(self) -> str:
        return await self._response.text()

    async def json(self) -> Any:
        return await self._response.json(
            loads=partial(json.loads, use_decimal=True, parse_constant=Decimal)
        )

    def raise_for_status(self) -> None:
        self._response.raise_for_status()


# Adds logging to aiohttp client session.
# https://stackoverflow.com/a/45590516/1466456
# Note that aiohttp client session is not meant to be extended.
# https://github.com/aio-libs/aiohttp/issues/3185
class ClientSession:
    def __init__(
        self,
        raise_for_status: Optional[bool] = None,
        name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        self._raise_for_status = raise_for_status
        self._name = name
        self._session = aiohttp.ClientSession(**kwargs)

    async def __aenter__(self) -> ClientSession:
        await self._session.__aenter__()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        await self._session.__aexit__(exc_type, exc, tb)

    def __del__(self, _warnings: Any = warnings) -> None:
        try:
            if not self._session.closed and self._name:
                _log.error(f"{self._name} unclosed client session.")
        except AttributeError:
            pass

    @asynccontextmanager
    async def request(
        self,
        method: str,
        url: str,
        name: Optional[str] = None,
        raise_for_status: Optional[bool] = None,
        **kwargs: Any,
    ) -> AsyncIterator[ClientResponse]:
        name = name or _rand_id()
        _log.info(f"Req {name} {method} {url}.")
        _log.debug(kwargs)
        async with self._session.request(method, url, **kwargs) as res:
            _log.info(f"Res {name} {res.status} {res.reason}.")
            content = {"headers": res.headers, "body": await res.text()}
            _log.debug(content)
            if raise_for_status or (raise_for_status is None and self._raise_for_status):
                res.raise_for_status()
            yield ClientResponse(res)

    @asynccontextmanager
    async def ws_connect(
        self, url: str, name: Optional[str] = None, **kwargs: Any
    ) -> AsyncIterator[ClientWebSocketResponse]:
        name = name or _rand_id()
        _log.info(f"Req {name} WS {url}: {kwargs}.")
        async with self._session.ws_connect(url, **kwargs) as ws:
            _log.info(f"Res {name} closed: {ws.closed} close code: {ws.close_code}.")
            yield ClientWebSocketResponse(ws, name)


class ClientWebSocketResponse:
    def __init__(self, client_ws_response: aiohttp.ClientWebSocketResponse, name: str) -> None:
        self._client_ws_response = client_ws_response
        self._name = name

    def __aiter__(self) -> ClientWebSocketResponse:
        return self

    async def __anext__(self) -> aiohttp.WSMessage:
        msg = await self._client_ws_response.__anext__()
        _log.debug(f"{self._name} {msg}.")
        return msg

    async def send_json(self, data: Any) -> None:
        _log.debug(f"{self._name} {data}.")
        await self._client_ws_response.send_json(data)

    async def close(self) -> None:
        await self._client_ws_response.close()

    async def receive(self) -> aiohttp.WSMessage:
        msg = await self._client_ws_response.receive()
        _log.debug(f"{self._name} {msg}")
        return msg


@asynccontextmanager
async def connect_refreshing_stream(
    session: ClientSession,
    url: str,
    interval: float,
    loads: Callable[[str], Any],
    take_until: Callable[[Any, Any], bool],
    name: Optional[str] = None,
    raise_on_disconnect: bool = False,
) -> AsyncIterator[AsyncIterable[Any]]:
    """
    Streams messages over WebSocket. The connection is restarted every `interval` seconds.
    Ensures no data is lost during restart when switching from one connection to another.
    """
    name = name or _rand_id()
    counter = cycle(range(0, 10))
    ctx, to_close_ctx = None, None
    timeout_task, receive_task = None, None

    async def inner() -> AsyncIterable[Any]:
        nonlocal ctx, to_close_ctx
        nonlocal timeout_task, receive_task
        assert ctx
        while True:
            to_close_ctx = None
            timeout_task = asyncio.create_task(asyncio.sleep(interval))
            while True:
                receive_task = asyncio.create_task(_receive(ctx.ws))
                done, _pending = await asyncio.wait(
                    (receive_task, timeout_task), return_when=asyncio.FIRST_COMPLETED
                )

                if timeout_task in done:
                    _log.info(f"Refreshing ws {ctx.name} connection.")
                    to_close_ctx = ctx
                    ctx = await _WSConnectionContext.connect(session, url, name, counter)

                    if receive_task.done():
                        new_msg = await _receive(ctx.ws)
                        assert new_msg.type is aiohttp.WSMsgType.TEXT
                        new_data = loads(new_msg.data)
                        async for old_msg in chain_async(
                            _resolved_stream(receive_task.result()), to_close_ctx.ws
                        ):
                            if old_msg.type is aiohttp.WSMsgType.CLOSED:
                                break
                            old_data = loads(old_msg.data)
                            if take_until(old_data, new_data):
                                yield old_data
                            else:
                                break
                        yield new_data

                    await to_close_ctx.close()
                    break

                msg = receive_task.result()
                if msg.type is aiohttp.WSMsgType.CLOSED:
                    if raise_on_disconnect:
                        _log.warning(
                            f"Server closed ws {ctx.name} connection; data: {msg.data}; raising "
                            "exception."
                        )
                        raise aiohttp.WebSocketError(
                            aiohttp.WSCloseCode.GOING_AWAY,
                            "Server unexpectedly closed WS connection.",
                        )
                    else:
                        _log.warning(
                            f"Server closed ws {ctx.name} connection; data: {msg.data}; "
                            "reconnecting."
                        )
                        timeout_task.cancel()
                        with suppress(asyncio.CancelledError):
                            await asyncio.gather(ctx.close(), timeout_task)
                        ctx = await _WSConnectionContext.connect(session, url, name, counter)
                        break

                yield loads(msg.data)

    try:
        ctx = await _WSConnectionContext.connect(session, url, name, counter)
        yield inner()
    finally:
        tasks: List[asyncio.Task] = []
        if receive_task:
            tasks.append(receive_task)
        if timeout_task:
            tasks.append(timeout_task)
        for task in tasks:
            task.cancel()
        with suppress(asyncio.CancelledError):
            await asyncio.gather(*tasks)
        if ctx:
            await ctx.close()
        if to_close_ctx:
            await to_close_ctx.close()


async def _receive(ws: ClientWebSocketResponse) -> aiohttp.WSMessage:
    while True:
        msg = await ws.receive()
        if msg.type in [aiohttp.WSMsgType.TEXT, aiohttp.WSMsgType.CLOSED]:
            return msg
        # Ping is handled implicitly by aiohttp because `autoping=True`. It will never reach here.
        # Close is handled implicitly by aiohttp because `autoclose=True`. It will reach here.
        if msg.type in [aiohttp.WSMsgType.CLOSING, aiohttp.WSMsgType.CLOSE]:
            continue
        raise NotImplementedError(f"Unhandled WS message. Type: {msg.type}; data: {msg.data}.")


class _WSConnectionContext:
    name: str
    conn: AsyncContextManager[ClientWebSocketResponse]
    ws: ClientWebSocketResponse

    @staticmethod
    async def connect(
        session: ClientSession, url: str, name: str, counter: Iterator[int]
    ) -> _WSConnectionContext:
        name = f"{name}-{next(counter)}"
        conn = session.ws_connect(url, name=name)
        ws = await conn.__aenter__()
        ctx = _WSConnectionContext()
        ctx.name = name
        ctx.conn = conn
        ctx.ws = ws
        return ctx

    async def close(self) -> None:
        await self.ws.close()
        await self.conn.__aexit__(None, None, None)


def _rand_id() -> str:
    return str(randint(100000, 999999))


async def _resolved_stream(*results: T) -> AsyncIterable[T]:
    for result in results:
        yield result
